use serde::{Serialize, Deserialize};
use std::collections::BTreeMap;
use crate::websocket;
use crate::server;

#[derive(Serialize, Deserialize, Clone, PartialEq, Debug)]
pub enum GameState {
    Starting,
    Creating,
    Guessing
}

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct CreatingEx {
    pub question: String,
    pub letters: Vec<char>
}

#[derive(Serialize, Clone, Debug)]
pub struct Player {
    pub nick: String,
    pub creating_exercise: Option<CreatingEx>,
    pub submitted_word: Option<String>,
    pub submitted_guess: Option<Vec<(String, String)>>
}

#[derive(Serialize, Clone, Debug)]
pub struct Game {
    pub players: Vec<Player>,
    pub round_number: i32,
    pub additional_questions: Vec<CreatingEx>,

    pub state: GameState,
}


impl Game {
    pub fn new() -> Self {
        Self {
            players: Vec::new(),
            round_number: 0,
            additional_questions: Vec::new(),
            state: GameState::Starting
        }
    }

    fn get_player(&mut self, nick: &str) -> Option<&mut Player> {
        self.players.iter_mut().find(|x| x.nick == nick)
    }

    pub fn check_letters(word: &str, letters: &Vec<char>) -> bool {
        let mut countmap: BTreeMap<char, isize> = BTreeMap::new();
        for c in letters {
            countmap.insert(*c, countmap.get(c).unwrap_or(&0) + 1);
        }
        for c in word.chars() {
            let count = countmap.get(&c);
            if let Some(&v) = count {
                if v < 0 {
                    return false;
                }
                countmap.insert(c, v - 1);
            }
            else {
                return false;
            }
        }

        true
    }

    pub fn submit_creation(&mut self, player_nick: &str, word: String) -> bool {
        match self.state {
            GameState::Creating => {
                if let Some(player) = self.get_player(player_nick) {
                    if let Some(ex) = &player.creating_exercise {
                        if Self::check_letters(&word, &ex.letters) {
                            player.submitted_word = Some(word);
                            return true;
                        }
                    }
                }
            },
            _ => {}
        }
        false
    }

    pub fn submit_guess(&mut self, player_nick: &str, guess: Vec<(String, String)>) -> bool {
        match self.state {
            GameState::Guessing => {
                if let Some(player) = self.get_player(player_nick) {
                    player.submitted_guess = Some(guess);
                    true
                }
                else {
                    false
                }
            },
            _ => false
        }
    }

    pub fn all_words_submitted(&self) -> bool {
        self.state == GameState::Creating &&
            self.players.iter().map(|p| p.submitted_word.is_some()).fold(true, |a, b| a && b)
    }
    pub fn all_guesses_submitted(&self) -> bool {
        self.state == GameState::Guessing &&
            self.players.iter().map(|p| p.submitted_guess.is_some()).fold(true, |a, b| a && b)
    }

    pub fn next_state(&mut self) {
        if self.state == GameState::Guessing {
            self.clear_submissions();
        }
        self.state = match self.state {
            GameState::Starting => GameState::Creating,
            GameState::Creating => GameState::Guessing,
            GameState::Guessing => GameState::Starting
        }
    }

    pub fn start_round(&mut self) {
        self.clear_submissions();

        let mut exes = vec![
            CreatingEx{ question: "Delikatessfutter für Hunde".to_owned(), letters: vec!['a', 'b', 'c'] },
            CreatingEx{ question: "A Betabet".to_owned(), letters: vec!['b', 'c', 'd'] },
            CreatingEx{ question: "A Gammaabet".to_owned(), letters: vec!['c', 'd', 'e'] },
            CreatingEx{ question: "A Deltaabet".to_owned(), letters: vec!['d', 'e', 'f'] },
        ];
        for p in &mut self.players {
            p.creating_exercise = exes.pop();
        }
        self.additional_questions.clear();
        self.additional_questions.append(&mut exes);
        self.state = GameState::Creating;
    }

    pub fn clear_submissions(&mut self) {
        for p in &mut self.players {
            p.submitted_word = None;
            p.submitted_guess = None;
            p.creating_exercise = None;
        }
    }

    pub fn player_join(&mut self, nick: String) {
        self.players.push(Player{
            nick: nick,
            creating_exercise: None,
            submitted_word: None,
            submitted_guess: None
        });
    }

    pub fn create_results(&self) -> Vec<Vec<Vec<String>>> {
        let mut result: Vec<Vec<Vec<String>>> = Vec::new();
        let mut questions = self.players.iter()
                        .map(|p| p.creating_exercise.clone().unwrap().question)
                        .collect::<Vec<_>>();
        questions.append(&mut self.additional_questions.iter()
                                .map(|x| x.question.clone())
                                .collect::<Vec<_>>()
                        );
        for player in &self.players {
            let mut line: Vec<Vec<String>> = Vec::new();
            let word = player.submitted_word.as_ref().unwrap();
            for question in &questions {
                let mut cell: Vec<String> = Vec::new();
                for p in &self.players {
                    if let Some(guess) = &p.submitted_guess {
                        if let Some(_corr) = guess.iter().find(|(w, q)| w == word && q == question) {
                            cell.push(p.nick.clone());
                        }
                    }
                }
                line.push(cell);
            }
            result.push(line);
        }
        return result;
    }
}

